'''
Computes and plots the frequency response of a 3d poroacoustic model presented
in R. Rumpler, P. Göransson, J.-F. Deü. "A finite element approach combining a
reduced-order system, Padé approximants, and an adaptive frequency windowing for
fast multi-frequency solution of poro-acoustic problems", Int. J. Numer. Methods
Eng. 97: 759--784, 2014.
For more information, see http://modelreduction.org/index.php/Porous_absorber.

Copyright 2023 Quirin Aumann

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

from time import time
from math import pi
from cmath import log10, sqrt

import h5py
from numpy import array, vdot
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spsolve, use_solver

import matplotlib.pyplot as plt

# Dimension of the matrices
n = 386076
m = 1
p = 1
fname = 'poroacoustic_n386076m1q1.mat'
use_solver(useUmfpack=True,assumeSortedIndices=True)

# Load data
with h5py.File(fname, 'r') as f:
    def read_data(var, dtype, n1=None, n2=None):
        if dtype == 'csc_matrix':
            return csc_matrix((f.get(var + '/data')[()], \
                f.get(var + '/ir')[()], \
                f.get(var + '/jc')[()]), \
                shape=(n1, n2))
        if dtype == 'array':
            return f.get(var)[()]
        if dtype == 'cplx_array':
            return f[var]['real'] + 1j*f[var]['imag']

    B = read_data('B', 'array').T
    C = read_data('C', 'array').T
    K = read_data('K', 'csc_matrix', n, n)
    K_p1 = read_data('K_p1', 'csc_matrix', n, n)
    K_p2 = read_data('K_p2', 'csc_matrix', n, n)
    M = read_data('M', 'csc_matrix', n, n)
    M_p1 = read_data('M_p1', 'csc_matrix', n, n)
    M_p2 = read_data('M_p2', 'csc_matrix', n, n)
    M_p3 = read_data('M_p3', 'csc_matrix', n, n)
    freq = read_data('f', 'array').flatten().tolist()
    mag = read_data('mag', 'array').flatten().tolist()

# Recompute result if required
RECOMPUTE = True
if RECOMPUTE:

    # Material parameters
    density_fluid           = 1.21
    viscosity_fluid         = 1.84e-5
    standard_pressure_fluid = 101e3
    heat_capacity_fluid     = 1.4
    prandtl_number_fluid    = 0.71
    porosity                = 0.96
    tortuosity              = 1.7
    flow_resistivity        = 32e3
    viscous_length          = 90e-6
    thermal_length          = 165e-6
    apparent_mass_density   = porosity * density_fluid * (tortuosity - 1)

    mag = []
    for i, f in enumerate(freq):
        print(f'Frequency step {i+1}, f={f:4.2f} Hz ... ', end='', flush=True)

        s = 1j*2*pi*f

        # Compute coefficients
        viscous_drag = flow_resistivity * porosity**2 * \
            sqrt(1 + 4 * s * tortuosity**2 * viscosity_fluid * density_fluid / \
                (flow_resistivity**2 * viscous_length**2 * porosity**2))
        alpha = 1 + 8 * viscosity_fluid / (s * prandtl_number_fluid * thermal_length**2 * density_fluid) * \
            sqrt(1 + s * prandtl_number_fluid * thermal_length**2 * density_fluid / (16 * viscosity_fluid))

        coeff_K_p1 = -porosity * (-apparent_mass_density - viscous_drag / s) / \
            (porosity * density_fluid + apparent_mass_density + viscous_drag / s)
        coeff_K_p2 = porosity**2 / (porosity * density_fluid + apparent_mass_density + viscous_drag / s)

        coeff_M_p1 = s**2 * (viscous_drag / s - (-apparent_mass_density - viscous_drag / s)**2 / \
            (porosity * density_fluid + apparent_mass_density + viscous_drag / s))
        coeff_M_p2 = s**2 * (-(heat_capacity_fluid-1) / (alpha * heat_capacity_fluid * standard_pressure_fluid))
        coeff_M_p3 = s**2 * coeff_K_p1

        # Assemble and solve system
        t = time()
        A = s**2 * M + K + coeff_K_p1 * K_p1 + coeff_K_p2 * K_p2 + \
            coeff_M_p1 * M_p1 + coeff_M_p2 * M_p2 + coeff_M_p3 * M_p3

        x = spsolve(A, (-s)**2 * B)

        mag.append(10 * log10(abs(C @ x) / 1e-9))

        print(f'finished in {time() - t:.2e} s.')

# Plot the results
plt.plot(freq,mag)
plt.xlim([100, 1000])
plt.xlabel('Frequency [Hz]')
plt.ylabel('Magnitude [dB]')
plt.show()

